///////////////////////////////////////////////////////////////////////////////
// File    : osx-mcp.m
// Content : multi-channel pointing interface on the mac
// History : 04-Mar-2009 - [rb] added CeCIIL licence, prepared distro
///////////////////////////////////////////////////////////////////////////////

/* licence ********************************************************************
Copyright Renaud Blanch [http://iihm.imag.fr/blanch/]

This software is a computer program whose purpose is to explore a pointing
technique coupling two input channels, movement and gaze, that control a grid
of cursor and which cursor is active; thus reduces drastically pointing time.

This software is governed by the CeCILL license under French law and abiding by 
the rules of distribution of free software. You can  use, modify and/ or 
redistribute the software under the terms of the CeCILL license as circulated  
at the following URL [http://www.cecill.info]. 

As a counterpart to the access to the source code and  rights to copy, modify 
and redistribute granted by the license, users are provided only with a limited 
warranty  and the software's author,  the holder of the economic rights,  and 
the successive licensors  have only limited liability. 

The fact that you are presently reading this means that you have had knowledge 
of the CeCILL license and that you accept its terms.
******************************************************************************/

/* globals *******************************************************************/

#import <Cocoa/Cocoa.h>

NSView *view;
static const double D = 250;


/* carbon callback ***********************************************************/

#include <Carbon/Carbon.h>

OSStatus mouse_handler(EventHandlerCallRef inRef, EventRef inEvent, void *userData) {
	[view setNeedsDisplay:true];
	return eventNotHandledErr;
}


/* view **********************************************************************/

@interface View : NSView {
	NSImage *arrow;
	NSInputStream *imouse;
	
	double dW, dH; // grid steps
	int Xg, Yg;    // gaze position
	int Xgm, Ygm;  // continuous gaze position
	
	int screen_width, screen_height;
	int width, height; // screen size in steps
}

@end


@implementation View

- (id)initWithFrame:(NSRect)frame 
{
	self = [super initWithFrame:frame];
	
	// instance
	view = self;
	EventTypeSpec events[] = { {kEventClassMouse, kEventMouseMoved},
	                           {kEventClassMouse, kEventMouseDragged} };
	InstallEventHandler(GetEventMonitorTarget(), NewEventHandlerUPP(mouse_handler),
	                    GetEventTypeCount(events), events, 0, 0);
			
	// constants
	dW = D * 1.;
	dH = D * sin(acos(-1)/3.);

	NSPoint mouse = [NSEvent mouseLocation];
	Xg = mouse.x;
	Yg = mouse.y;
	
	NSSize size = frame.size;
	screen_width = size.width;
	screen_height = size.height;
	width = (screen_width / dW) + 2;
	height = (screen_height / dH) + 1;

	// arrow image
	arrow = [[NSImage alloc] initWithContentsOfFile:@"./arrow.png"];

	// mouse input stream
	imouse = [[NSInputStream alloc] init];
	NSHost *host = [NSHost hostWithName:@"localhost"];
	[NSStream getStreamsToHost:host port:1235 inputStream:&imouse outputStream:nil];
	[imouse setDelegate:self];
	[imouse open];
	[imouse scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
	
	return self;
}

- (void)stream:(NSStream *)stream handleEvent:(NSStreamEvent)eventCode
{
	switch(eventCode) {
		case NSStreamEventHasBytesAvailable: 
		{
			static const int MAX_LEN = 1024;
			char buf[MAX_LEN];
			unsigned int len = [(NSInputStream *)stream read:(uint8_t *)buf maxLength:MAX_LEN];
			if(len > 0) {
				sscanf(buf, "g %i %i", &Xg, &Yg);
				Yg = screen_height - Yg;
//				NSLog(@"g %i %i", Xg, Yg);
				Xgm = Xg; Ygm = Yg;
			}
			break;
		}
	}
}

- (void)drawRect:(NSRect)rect 
{
	// gaze
/*
	static const int GAZE_SIZE = 40;
	NSPoint gaze_position = { Xg - GAZE_SIZE/2, Yg - GAZE_SIZE/2 };
	NSRect gaze_rect;
	gaze_rect.origin = gaze_position;
	gaze_rect.size.width = GAZE_SIZE;
	gaze_rect.size.height = GAZE_SIZE;
	NSBezierPath *path = [NSBezierPath bezierPathWithRoundedRect:gaze_rect xRadius:GAZE_SIZE yRadius:GAZE_SIZE];
	NSColor *red = [NSColor colorWithDeviceRed:1. green:0. blue:0. alpha:.6];
	[red set];
	[path fill];
*/

	NSSize size = [arrow size];
	size.width *= 2;
	size.height *= 2;
	
	// mouse
	NSPoint mouse = [NSEvent mouseLocation];
	int Xm = mouse.x;
	int Ym = mouse.y;
//	NSLog(@"m %i %i", Xm, Ym);
		
	double d_min = screen_width + screen_height;
	int Xb, Yb; // best cursor position
	
	int i0 = -(mouse.x/dW + .5);
	int j0 = -(mouse.y/dH);
	int i, j;
	for(i = i0; i < i0 + width; i += 1) {
		for(j = j0; j < j0 + height; j += 1) {
			int X = Xm + i*dW + (j%2) * dW/2;
			int Y = Ym + j*dH;
			if(((-12 <= X) && (X <= screen_width) && (0 <= Y) && (Y <= screen_height + 22))) {
				double d = hypot(X - Xgm, Y - Ygm);
				if(d < d_min) {
					d_min = d;
					Xb = X;
					Yb = Y;
				}
				if(i != 0 || j != 0) {
//					NSRect dest = { NSMakePoint(X - (4*2), Y - (19*2)), size };
//					[arrow drawInRect:dest fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:.6];
					[arrow compositeToPoint:NSMakePoint(X - 4, Y - 19) operation:NSCompositeSourceOver fraction:.4];
				}
			}
		}
	}
	if(Xb != Xm || Yb != Ym) {
		CGPoint warp_mouse = { Xb, screen_height - Yb };
		CGEventPost(kCGHIDEventTap, CGEventCreateMouseEvent(NULL, kCGEventMouseMoved, warp_mouse, 0));
	}
	
	Xgm = Xb; // to ensure continuity between gaze
	Ygm = Yb;	
}

@end



/* main **********************************************************************/

int main(int argc, char *argv[]) {
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	NSApplication *app = [NSApplication sharedApplication];

	NSWindow *window = [[NSWindow alloc] 
	                    initWithContentRect:[[NSScreen mainScreen] frame]
	                    styleMask:NSBorderlessWindowMask
	                    backing:NSBackingStoreBuffered
	                    defer:NO];
	[window setLevel:CGShieldingWindowLevel()];
	[window setIgnoresMouseEvents:YES];
	[window setBackgroundColor:[NSColor clearColor]];
	[window setOpaque:NO];
	
	view = [[View alloc] initWithFrame:[[NSScreen mainScreen] frame]];
	[window setContentView:view];
	[window setDelegate:view];

	[window makeKeyAndOrderFront: nil];


	[app run];

	
	[app release];
	[pool release];
	
	return EXIT_SUCCESS;
}